# Copyright 2011 Terena. All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:

#    1. Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.

#    2. Redistributions in binary form must reproduce the above copyright notice,
#       this list of conditions and the following disclaimer in the documentation
#        and/or other materials provided with the distribution.

# THIS SOFTWARE IS PROVIDED BY TERENA ``AS IS'' AND ANY EXPRESS OR IMPLIED
# WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
# EVENT SHALL TERENA OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
# OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# The views and conclusions contained in the software and documentation are those
# of the authors and should not be interpreted as representing official policies,
# either expressed or implied, of Terena.
import uuid

from django.conf import settings
from django.utils.importlib import import_module
from django.db.models.fields.files import FieldFile, FileField

from vff.storage import VersionedStorage
from vff.abcs import VFFBackend

HAS_SOUTH = True
try:
    from south.modelsinspector import add_introspection_rules
except ImportError:
    HAS_SOUTH = False


class VersionedFieldFile(FieldFile):

    def save(self, name, content, username='', commit_msg='', save=True):
        if not username:
            return
        if self.instance.pk is None:    # new file
            self.name = uuid.uuid4().hex
        else:
            self.name = self.storage.backend.get_media_path(self.instance)
            save = False
        self.storage.save(self.name, content, username, commit_msg, save)
        setattr(self.instance, self.field.name, self.name)

        # Update the filesize cache
        self._size = content.size
        self._committed = True
    save.alters_data = True

    def delete(self, username='', commit_msg='', save=False):
        if not username:
            return
        if hasattr(self, '_file'):
            self.close()
            del self.file
        self.storage.delete(self.name, username, commit_msg, save)

    delete.alters_data = True

    def list_revisions(self, count=0, offset=0):
        return self.storage.backend.list_revisions(self.instance,
                                           count=count, offset=offset)

    def get_revision(self, rev=None):
        return self.storage.backend.get_revision(self.instance, rev=rev)

    def get_diff(self, r1, r2):
        return self.storage.backend.get_diff(self.instance, r1, r2)


class VersionedFileField(FileField):

    attr_class = VersionedFieldFile

    def __init__(self, name=None, verbose_name=None, storage=None, **kwargs):
        try:
            path = settings.VERSIONEDFILE_BACKEND
        except AttributeError:
            raise NameError('When using VersionedField, you have to define'
                            ' VERSIONEDFILE_BACKEND in settings.py. Refer'
                            ' to the docs for more info.')
        mname = '.'.join(path.split('.')[:-1])
        cname = path.split('.')[-1]
        module = import_module(mname)
        backend_class = getattr(module, cname)
        if not issubclass(backend_class, VFFBackend):
            raise ValueError('The class pointed at in VERSIONEDFILE_BACKEND'
                             ' has to provide the interface defined by'
                             ' vff.abcs.VFFBackend.')
        vstorage = VersionedStorage(backend_class, name)
        super(VersionedFileField, self).__init__(verbose_name=verbose_name,
                                                 name=name,
                                                 upload_to='unused',
                                                 storage=vstorage,
                                                 **kwargs)

if HAS_SOUTH:
    add_introspection_rules([
        (
            [VersionedFileField],
            [],
            {},
        ),
    ], ["^vff\.field\.VersionedFileField"])



# Copyright 2011 Terena. All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:

#    1. Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.

#    2. Redistributions in binary form must reproduce the above copyright notice,
#       this list of conditions and the following disclaimer in the documentation
#        and/or other materials provided with the distribution.

# THIS SOFTWARE IS PROVIDED BY TERENA ``AS IS'' AND ANY EXPRESS OR IMPLIED
# WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
# EVENT SHALL TERENA OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
# OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# The views and conclusions contained in the software and documentation are those
# of the authors and should not be interpreted as representing official policies,
# either expressed or implied, of Terena.

import os
import re
import datetime
import difflib
import git
from types import MethodType
from tempfile import NamedTemporaryFile

from django.conf import settings
from django.core.files.move import file_move_safe

from vff.abcs import VFFBackend

USERPAT = re.compile(ur'^([^<]+) <(.+)>$')
EMAILPAT = re.compile(ur'^([^@]+)@.+$')

GITCONFIG = u'''\
[user]
 name = %s
 email = %s
'''

def clean_environment():
    for env in ('USER', 'USERNAME',
            'GIT_AUTHOR_NAME', 'GIT_AUTHOR_EMAIL',
            'GIT_COMMITTER_NAME', 'GIT_COMMITTER_EMAIL',
            'GIT_AUTHOR_DATE', 'GIT_COMMITER_DATE'):
        if env in os.environ:
            del os.environ[env]
    os.environ['USERNAME'] = 'dummy@dummy'

def get_repo_name():
    """
    get the name of the versioned files repository
    """
    try:
        reponame = settings.VFF_REPO_NAME
    except AttributeError:
        reponame = 'vf_repo'
    return reponame

def get_repo_location():
    """
    get the absolute path and the base_url for the
    versioned files repository
    """
    reponame = get_repo_name()
    location = os.path.join(settings.MEDIA_ROOT, reponame)
    return os.path.abspath(location)

def create_fname(instance, fieldname):
    """
    return the path to the file relative to the
    repository of versioned files
    """
    class_name = instance.__class__.__name__.lower()
    return '%s%s-%s.xml' % (class_name, instance.pk, fieldname)


class Repo(git.Repo):
    """
    This class is only to get rid of the __slots__
    nuisance in the original class, whereupon you cannot
    override instance methods.
    """


class GitBackend(object):
    """
    Git backend for versioned file field's storage.
    See abcs.py for documentation.
    """

    def __init__(self, fieldname):
        self.location = get_repo_location()
        self.fieldname = fieldname
        try:
            self.repo = Repo(self.location)
        except git.exc.NoSuchPathError:
            self.repo = Repo.init(self.location)
            readme = os.path.join(self.location, 'README')
            f = open(readme, 'w')
            f.write('VFF GIT REPOSITORY')
            f.close()
            self.repo.index.add([readme])
            self.repo.index.commit('Initial commit')

    def get_media_path(self, instance):
        return os.path.join(get_repo_name(),
                create_fname(instance, self.fieldname))

    def _commit(self, fname, msg, username, action):
        mu = USERPAT.match(username)
        me = EMAILPAT.match(username)
        if mu:
            config = GITCONFIG % (mu.group(1), mu.group(2))
        elif me:
            config = GITCONFIG % (me.group(1), me.group(0))
        else:
            config = GITCONFIG % (username, username)
        with NamedTemporaryFile(delete=True) as f:
            f.write(config.encode('utf8'))
            f.seek(0)
            def fun(self, config_level=None):
                return f
            meth = MethodType(fun, self.repo, Repo)
            setattr(self.repo, '_get_config_path', meth)
            setattr(self.repo, 'config_level', ['repository'])
            clean_environment()
            if action == 'add':
                self.repo.index.add([fname])
            elif action == 'delete':
                try:
                    self.repo.index.remove([fname], working_tree=True)
                except git.GitCommandError:
                    pass
            self.repo.index.commit(msg)

    def add_revision(self, content, instance, commit_msg, username):
        fname = create_fname(instance, self.fieldname)
        full_path = os.path.join(self.location, fname)
        if hasattr(content, 'temporary_file_path'):
            # This file has a file path that we can move.
            file_move_safe(content.temporary_file_path(), full_path)
            content.close()
        else:
            # This is a normal uploadedfile that we can stream.
            with open(full_path, 'w') as f:
                content.seek(0)
                f.write(content.read())
        if settings.FILE_UPLOAD_PERMISSIONS is not None:
            os.chmod(full_path, settings.FILE_UPLOAD_PERMISSIONS)
        self._commit(fname, commit_msg, username, 'add')

    def del_document(self, instance, commit_msg, username):
        fname = create_fname(instance, self.fieldname)
        self._commit(fname, commit_msg, username, 'delete')

    def list_revisions(self, instance, count=0, offset=0):
        fname = create_fname(instance, self.fieldname)
        revs = []
        kwargs = {}
        if count:
            kwargs['count'] = count
        if offset:
            kwargs['offset'] = offset
        for ci in self.repo.iter_commits(paths=fname, **kwargs):
            rev = {'versionid': ci.hexsha,
                   'author': ci.author.name,
                   'message': ci.message,
                   'date': datetime.datetime.fromtimestamp(ci.committed_date),}
            revs.append(rev)
        return revs

    def get_revision(self, instance, rev=None):
        fname = create_fname(instance, self.fieldname)
        full_path = os.path.join(self.location, fname)
        text = u''
        if rev:
            blob = self.repo.commit(rev).tree[fname]
            text = blob.data_stream[3].read()
        elif os.path.exists(full_path):
            with open(full_path) as f:
                text = f.read()
        return text.decode('utf8')
    
    def get_diff(self, instance, r1, r2):
        md1 = self.get_revision(instance, r1).split(u'\n')
        md2 = self.get_revision(instance, r2).split(u'\n')
        diff = u'\n'.join(difflib.unified_diff(md1, md2,
                                              fromfile=r1,
                                              tofile=r2,
                                              ))
        return diff

VFFBackend.register(GitBackend)



# Copyright 2011 Terena. All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:

#    1. Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.

#    2. Redistributions in binary form must reproduce the above copyright notice,
#       this list of conditions and the following disclaimer in the documentation
#        and/or other materials provided with the distribution.

# THIS SOFTWARE IS PROVIDED BY TERENA ``AS IS'' AND ANY EXPRESS OR IMPLIED
# WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
# EVENT SHALL TERENA OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
# OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# The views and conclusions contained in the software and documentation are those
# of the authors and should not be interpreted as representing official policies,
# either expressed or implied, of Terena.

import os
from setuptools import setup, find_packages


def read(*rnames):
    return open(os.path.join(os.path.dirname(__file__), *rnames)).read()

version = '0.1b3'

setup(name='django-vff',
      version=version,
      description=("Versioned file field for django models"),
      long_description=(read('README.rst') + '\n\n' + read('CHANGES.rst')),
      classifiers=[
        'Development Status :: 4 - Beta',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',],
      keywords='django versioned file field vcs model git',
      author='TERENA',
      author_email='eperez@yaco.es',
      url='https://github.com/Yaco-Sistemas/django-vff',
      license='BSD',
      packages=find_packages(),
      include_package_data=True,
      zip_safe=False,
      install_requires=[
          'GitPython',
      ],
)



# Copyright 2011 Terena. All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:

#    1. Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.

#    2. Redistributions in binary form must reproduce the above copyright notice,
#       this list of conditions and the following disclaimer in the documentation
#        and/or other materials provided with the distribution.

# THIS SOFTWARE IS PROVIDED BY TERENA ``AS IS'' AND ANY EXPRESS OR IMPLIED
# WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
# EVENT SHALL TERENA OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
# OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# The views and conclusions contained in the software and documentation are those
# of the authors and should not be interpreted as representing official policies,
# either expressed or implied, of Terena.

import os

from django.conf import settings
from django.db.models.signals import post_save, post_delete
from django.core.files.storage import FileSystemStorage
from django.utils.encoding import force_unicode


class VersionedStorage(FileSystemStorage):
    """
    Versioned filesystem storage
    """

    def __init__(self, backend_class, fieldname):
        self.backend = backend_class(fieldname)
        self.location = os.path.abspath(settings.MEDIA_ROOT)
        self.fieldname = fieldname

    def save(self, uid, content, username, commit_msg, save):
        def savefile(sender, instance=None, **kwargs):
            # check that the instance is the right one
            try:
                saved_uid = getattr(instance, self.fieldname).name
            except AttributeError:
                # an instance of another class
                return
            if saved_uid != uid:
                return
            # create the actual filename from the versioned file
            name = self.backend.get_media_path(instance)
            content.name = name
            # new revision
            self.backend.add_revision(content, instance, commit_msg, username)

            if save or kwargs['created']:
                setattr(instance, self.fieldname, name)
                instance.save()

            # remove signal
            post_save.disconnect(dispatch_uid=uid)

        post_save.connect(savefile, weak=False, dispatch_uid=uid)
        return force_unicode(uid.replace('\\', '/'))

    def delete(self, uid, username, commit_msg, save):
        def deletefile(sender, instance=None, **kwargs):
            # check that the instance is the right one
            fieldfile = getattr(instance, self.fieldname)
            try:
                saved_uid = fieldfile.name
            except AttributeError:
                # an instance of another class
                return
            if saved_uid != uid:
                return
            fieldfile.name = None
            setattr(instance, fieldfile.field.name, fieldfile.name)
            self.backend.del_document(instance, commit_msg, username)

            # Delete the filesize cache
            if hasattr(fieldfile, '_size'):
                del fieldfile._size
            fieldfile._committed = False

            if save:
                instance.save()

            # remove signal
            post_delete.disconnect(dispatch_uid=uid)

        post_delete.connect(deletefile, weak=False, dispatch_uid=uid)
        return force_unicode(uid.replace('\\', '/'))



# Copyright 2011 Terena. All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:

#    1. Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.

#    2. Redistributions in binary form must reproduce the above copyright notice,
#       this list of conditions and the following disclaimer in the documentation
#        and/or other materials provided with the distribution.

# THIS SOFTWARE IS PROVIDED BY TERENA ``AS IS'' AND ANY EXPRESS OR IMPLIED
# WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
# EVENT SHALL TERENA OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
# OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# The views and conclusions contained in the software and documentation are those
# of the authors and should not be interpreted as representing official policies,
# either expressed or implied, of Terena.

from abc import ABCMeta, abstractmethod


class VFFBackend(object):

    __metaclass__ = ABCMeta

    @abstractmethod
    def __init__(self, fieldname):
        """
        Initialize the backend.

        params:
        - fieldname: a string with the name of the field that corresponds to
                    this storage backend
        """

    @abstractmethod
    def get_media_path(self, instance):
        """
        Get the path to the corresponding file relative
        to the django media directory

        params:
        - instance: The django model object corresponding to this content
        """

    @abstractmethod
    def add_revision(self, content, instance, commit_msg, username):
        """
        Add a new revision to an existing document, or add a new document
        to the repository.

        params:
        - content: A file like object that implements open, seek, read, close
                  and contains the document data to be versioned
        - instance: The django model object corresponding to this content
        - commit_msg: A string with the commit msg
        - username: A username to commit with
        """

    @abstractmethod
    def del_document(self, instance, commit_msg):
        """
        Remove document from the repository.

        params:
        - instance: The django model object corresponding to this content
        - commit_msg: A string with the commit msg
        """

    @abstractmethod
    def list_revisions(self, instance, count=0, offset=0):
        """
        return a list of all (or from offset to offset+count) revisions of a document. A revision is here
        represented by a dictionary with keys:
            * versionid: A string that uniquelly identifies a version for the document, and can be used
                         as parameter for self.get_revision (below)
            * author: A userid
            * message: A string with the commit message
            * date: a datetime object with the date the revision was commited.

        params:
        - instance: The django model object corresponding to this content
        - count: The number of revisions to list. -1 for all.
        - offset: The number of revisions skipped from the list
        """

    @abstractmethod
    def get_revision(self, instance, rev=None):
        """
        return the revision specified by id. If id is None, the last revision.
        A revision should be a string encoded as utf8.

        params:
        - instance: The django model object corresponding to this content
        - rev: the id of the revision to get. If None, get the last.
        """

    @abstractmethod
    def get_diff(self, instance, id1, id2):
        """
        return a diff between two revisions, as an utf8 string.

        params:
        - instance: The django model object corresponding to this content
        - id1, id2: the ids of the revisions to diff.
        """
